/**
 * $Id: Main.java,v 1.2 2001/07/27 02:06:59 pcj Exp $
 *
 * Copyright (C) 2001 Paul Cody Johnston - pcj@inxar.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA.
 **/
package example;

import java.util.Date;
import org.inxar.hotswap.*;

/**
 * <code>Main</code> starts a <code>Thread</code> that periodically
 * wakes up and calls <code>Example e =
 * (Example)proxy.hotswap()</code> on a <code>Proxy</code> instance
 * that is constructed from the classname you specify.
**/
public class Main 
{
    private Main()
    {
    }

    // ================================
    // Class Methods
    // ================================

    /**
     * <code>Main</code> does not do any custom
     * <code>ProxyCompiler</code> configuration.  It is expected that
     * a configuration file will be passed to the application via the
     * -D option.  See the manual for more information.
     * 
     * Usage: 
     * <pre>
     * java example.Main &lt;Example implementation classname&gt;
     * </pre>
     *
    **/
    public static void main(String[] argv) throws ProxyException
    {
	if (argv.length != 1) {
	    System.err.println("Usage: java example.Main "+
			       "<Example implementation classname>");
	    System.exit(1);
	}

	String className = argv[0];
	String version = System.getProperty("java.version");
	if (version.startsWith("1.3"))
	    isJDK13 = true;

	// Setup ProxyClassLoader
	ProxyClassLoader loader = new ProxyClassLoader();

	// Setup ProxyClass
	ProxyClass cls =  (isJDK13) 
	    ? loader.load(className)
	    : loader.loadJDK12(className);
	
	// Setup the Proxy
	Proxy proxy = cls.newInstance();

	// Setup an event listener
	ProxyEventListener l = new ProxyEventListener() {
		public void notify(ProxyEvent evt) {
		    Main.trace(evt.toString());
		}
	    };
	cls.addClassSwapListener(l);
	proxy.hotswap_addObjectSwapListener(l);
	
	// Start the hotswap thread
	log = loader.getLog();
	channel = log.newChannel("run", new Main());
	trace("Starting HotSwap thread on "+className+"...");
	Thread t = new Thread(new HotSwapper(proxy));
	t.start();
    }
    
    static void trace(String msg)
    {
	if (channel != null) 
	    channel.info(msg);
	else 
	    System.out.println("[Main]["+new Date()+"] "+msg);
    }

    static void logoff()
    {
	if (log != null) 
	    if (log instanceof ConsoleLog)
		((ConsoleLog)log).close();
    }

    // ================================
    // Class Fields
    // ================================

    static boolean isJDK13;
    static ProxyLog.Channel channel;
    static ProxyLog log;

    // ================================
    // Inner Classes
    // ================================

    static class HotSwapper implements Runnable
    {
	HotSwapper(Proxy proxy) 
	{
	    this.proxy = proxy;
	}

	public void run()
	{
	    try {

		while (true) {
		
		    Example example = proxy.hotswap_getProxyClass().isJDK13()
			? (Example)proxy
			: (Example)proxy.hotswap();
		
		    trace("Calling example.execute() now...");
		    example.execute();

		    long sleep = example.getSleepMillis();
		    if (sleep < 0)
			throw new InterruptedException();

		    trace("Sleeping for "+sleep+" ms...");
		    Thread.currentThread().sleep(sleep);;
		}

	    } catch (InterruptedException iex) {
		trace("Interrupted! Goodbye...");
	    } finally {
		logoff();
	    }
	}

	Proxy proxy;
    }
}
