/**
 * $Id: ConsoleLog.java,v 1.2 2001/07/27 06:44:07 pcj Exp $
 *
 * Copyright (C) 2001 Paul Cody Johnston - pcj@inxar.org
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */
package org.inxar.hotswap;

import java.io.Writer;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.io.BufferedWriter;
import java.io.OutputStreamWriter;
import java.io.IOException;

import java.util.Map;
import java.util.TreeMap;
import java.util.Iterator;
import java.util.Properties;

/**
 * Concrete implementation of <code>ProxyLog</code> that writes to the
 * console (or any <code>Writer</code>, to be precise).  The logging
 * priorities can be set by OR'ing the PRI_XXX constants and passing
 * it to the appropriate constructor.  The user should also call
 * <code>close()</code> at the end of the session.
 *
 * @since 0.8.6
 **/
public class ConsoleLog implements ProxyLog
{
    private static final boolean DEBUG = false;

    // ================================
    // Constants 
    // ================================

    /**
     * Logging priority constant.
    **/
    public static final int PRI_DEBUG = 0x0000000F;

    /**
     * Logging priority constant.
    **/
    public static final int PRI_INFO  = 0x000000F0;

    /**
     * Logging priority constant.
    **/
    public static final int PRI_WARN  = 0x00000F00;

    /**
     * Logging priority constant.
    **/
    public static final int PRI_FATAL = 0x0000F000;

    // ================================
    // Constructors 
    // ================================

    /**
     * Constructs a <code>ConsoleLog</code> that writes to
     * <code>System.out</code> with all priorities enabled.
     **/
    public ConsoleLog()
    {
	this(null);
    }

    /**
     * Constructs a <code>ConsoleLog</code> that writes to the given
     * <code>Writer</code> with all priorites enabled.
     **/
    public ConsoleLog(Writer out)
    {
	this(PRI_INFO | PRI_WARN | PRI_FATAL | PRI_DEBUG, out);
    }

    /**
     * Constructs a <code>ConsoleLog</code> that writes to
     * <code>System.out</code> with the given priorities enabled.
     **/
    public ConsoleLog(int priorities)
    {
	this(priorities, null);
    }

    /**
     * Constructs a <code>ConsoleLog</code> that writes to the given
     * <code>Writer</code> with the given priorities enabled.
     **/
    public ConsoleLog(int priorities, Writer out) 
    {
	this.priorities = priorities;
	this.channels = new TreeMap();
	
	if (out == null)
	    this.out = new BufferedWriter
		(new OutputStreamWriter(System.out));	
	else if (!(out instanceof BufferedWriter))
	    this.out = new BufferedWriter(out);
	else
	    this.out = out;

	if ((priorities & PRI_INFO) != 0) {
	    KBuffer buf = new KBuffer()
		.add("Logging session initiated at ")
		.add(new java.util.Date().toString())
		.ln()
		.add("---------------------------------------------")
		.ln();

	    try {
		buf.toWriter(this.out);
	    } catch (IOException ioex) {
		ioex.printStackTrace();
	    } 
	}
    }

    // ================================
    // TECH Methods 
    // ================================

    public String toString()
    {
	StringBuffer b = new StringBuffer()
	    .append(this.getClass().getName())
	    .append(" (FWID=")
	    .append((priorities & PRI_FATAL) != 0 ? 't' : 'f')
	    .append((priorities & PRI_WARN)  != 0 ? 't' : 'f')
	    .append((priorities & PRI_INFO)  != 0 ? 't' : 'f')
	    .append((priorities & PRI_DEBUG) != 0 ? 't' : 'f')
	    .append(')');
	return b.toString();
    }

    // ================================
    // ProxyLog Methods 
    // ================================

    public synchronized ProxyLog.Channel newChannel(String name, Object obj)
    {
	KChannel channel = (KChannel)channels.get(name);

	if (channel == null) {
	    channel = this.new KChannel(name, obj, 1);
	    channels.put(name, channel);
	} else {
	    channel = this.new KChannel(name, obj, ++channel.num);
	}

	return channel;
    }

    /**
     * This is used within ProxyClassLoader.initProxyLog
    **/
    public boolean initialize(Properties p)
    {
	this.priorities = 0;
	initPriority(p, "log.fatal", PRI_FATAL);
	initPriority(p, "log.warn",  PRI_WARN);
	initPriority(p, "log.info",  PRI_INFO);
	initPriority(p, "log.debug", PRI_DEBUG);
	return true;
    }

    // ================================
    // Other Methods 
    // ================================

    protected void initPriority(Properties p, String k, int t)
    {
	String v = ProxyProperties.getFirstProperty(k, p);
	if (v != null) {
	    try {
		if (new Boolean(v).booleanValue())
		    addPriority(t);
		else 
		    removePriority(t);
	    } catch (RuntimeException ex) {
		ex.printStackTrace();
	    } 
	}
    }

    /**
     * OR's the given priority level to the current set of activated
     * log priority levels.
    **/
    public void addPriority(int priority)
    {
	this.priorities |= priority;
    }

    /**
     * AND NOT's the given priority level from the current set of
     * activated log priority levels.
    **/
    public void removePriority(int priority)
    {
	this.priorities &= ~priority;
    }

    /**
     * Sets the given priority levels.
    **/
    public void setPriority(int priorities)
    {
	this.priorities = priorities;
    }

    /**
     * Closes the Stream was well as printing a summary of the
     * channels that were involved in the logging session.
    **/
    public synchronized void close()
    {
	if (out != null) {

	    try {
	       
		if ((priorities & PRI_INFO) != 0) {
		    KBuffer buf = new KBuffer()
			.add("---------------------------------------------").ln();
		    
		    Iterator i = channels.values().iterator();
		    KChannel c;
		    while (i.hasNext()) {
			c = (KChannel)i.next();
			buf.add('"')
			    .add(c.name)
			    .add(c.num)
			    .add('"')
			    .add(" was ")
			    .add(c.obj.getClass().getName())
			    .add('.')
			    .add(System.identityHashCode(c.obj))
			    .ln();
		    }
		    
		    buf.add("Logging session ended at ")
			.add(new java.util.Date().toString())
			.ln();
		    
		    buf.toWriter(out);
		}

		synchronized (out) {
		    out.close(); 
		    out = null;
		}

	    } catch (IOException ioex) {
		if (out != null)
		    try { out.close(); out = null; } 
		    catch (Exception ex) {}
		ioex.printStackTrace();
	    }

	}
    }

    int lineNum()
    {
	return ++lineNum;
    }

    static int encode(Channel c)
    {
	int lp = 0;
	if (c.doDebug()) lp |= PRI_DEBUG;
	if (c.doInfo())  lp |= PRI_INFO;
	if (c.doWarn())  lp |= PRI_WARN;
	if (c.doFatal()) lp |= PRI_FATAL;
	return lp;
    }

    // ================================
    // Fields 
    // ================================

    protected int lineNum;	// current line number
    protected int priorities;	// bitset of PRI_XXX's
    protected Map channels;	// map of channels by name
    protected Writer out;	// the charsink

    // ================================
    // Inner Classes 
    // ================================

    protected class KChannel implements ProxyLog.Channel
    {
	KChannel(String name, Object obj, int num)
	{
	    this.name = name; 
	    this.obj = obj;
	    this.num = num;
	    this.buf = new KBuffer();
	}

	public String toString()
	{
	    return name + num + '.' + System.identityHashCode(this);
	}

	public boolean doDebug()
	{
	    return (PRI_DEBUG & priorities) != 0;
	}

	public boolean doInfo()
	{
	    return (PRI_INFO & priorities) != 0;
	}

	public boolean doWarn()
	{
	    return (PRI_WARN & priorities) != 0;
	}

	public boolean doFatal()
	{
	    return (PRI_FATAL & priorities) != 0;
	}

	public void debug(Object msg)
	{
	    if (doDebug())
		write("debug", msg);
	}

	public void info(Object msg)
	{
	    if (doInfo())
		write("info", msg);
	}

	public void warn(Object msg)
	{
	    if (doWarn())
		write("warn", msg);
	}

	public void warn(Object msg, Throwable t)
	{
	    if (doWarn())
		write("warn", msg, t);
	}

	public void fatal(Object msg)
	{
	    if (doFatal())
		write("fatal", msg);
	}

	public void fatal(Object msg, Throwable t)
	{
	    if (doFatal())
		write("fatal", msg, t);
	}

	protected void write(String pri, Object msg)
	{
	    buf.add('[').add(lineNum()).add(']');
	    buf.add('[').add(name).add(num).add(']');
	    buf.add('[').add(pri).add(']');
	    buf.add(' ').add(msg.toString());
	    buf.ln();

	    if (out != null)
		try {
		    buf.toWriter(out);
		    synchronized (out) {
			out.flush();
		    }
		} catch (IOException ioex) {
		    ioex.printStackTrace();
		} 
	    else 
		System.out.println(buf.toString());

	    buf.length(0);
	}

	protected void write(String pri, Object msg, Throwable t)
	{
	    if (msg != null && t != null) {
		StringWriter out = new StringWriter();
		t.printStackTrace(new PrintWriter(out));
		msg = msg.toString() + ':' + out.toString();
	    }

	    write(pri, msg);
	}

	String name;		// name of this channel
	int num;		// # of channels with this name
	Object obj;		// the channel for this object
	KBuffer buf;		// stringbuffer
    }
    
}

// ================================
// Other Classes 
// ================================

/**
 * StringBuffer that renders to a Writer.
**/
class KBuffer implements Cloneable
{
    private static final String NEWLINE = System.getProperty("line.separator");
    private static final boolean DEBUG = false;

    public KBuffer()
    {
	this(17);
    }
    
    public KBuffer(int capacity)
    {
	this.src = new char[capacity];
	this.len = 0;
    }
    
    public KBuffer set(int offset, String s)
    {
	return set(offset, s.toCharArray(), 0, s.length());
    }

    public KBuffer set(int offset, char[] that_src, int that_off, int length)
    {

	while (this.src.length < offset + length)
	    realloc();
	
	System.arraycopy(that_src, that_off, this.src, offset, length);

	this.len = Math.max(this.len, offset + length);

	return this;
    }

    public KBuffer add(String val, int offset, int length)
    {
	if (val == null) 
	    val = String.valueOf(val);

	while (this.src.length < this.len + length)
	    realloc();
	
	if (DEBUG)
	    System.out.println("Copying from string with length "+val.length()+
			       " starting at "+offset+
			       " and ending at "+(offset+length)+
			       " into array with length "+this.src.length+
			       " starting at "+this.len);

	val.getChars(offset, offset + length, this.src, this.len);

	this.len += length;

	return this;
    }

    public KBuffer add(String val)
    {
	if (val == null) 
	    val = "null";

	return add(val, 0, val.length());
    }

    public KBuffer add(boolean val)
    {
	return add(String.valueOf(val));
    }

    public KBuffer add(char val)
    {
	return add(String.valueOf(val));
    }

    public KBuffer add(int val)
    {
	return add(String.valueOf(val));
    }

    public KBuffer add(long val)
    {
	return add(String.valueOf(val));
    }

    public KBuffer ln()
    {
	return add(NEWLINE);
    }

    public KBuffer add(KBuffer that)
    {
	while (this.src.length < this.len + that.len)
	    realloc();
	
	System.arraycopy(that.src, 0, this.src, this.len, that.len);

	this.len += that.len;

	return this;
    }

    public KBuffer add(char[] that_src)
    {
	return add(that_src, 0, that_src.length);
    }

    public KBuffer add(char[] that_src, int that_off, int that_len)
    {
	while (this.len + that_len > this.src.length)
	    realloc();

	System.arraycopy(that_src, that_off, this.src, this.len, that_len);

	return this;
    }

    public KBuffer setCharAt(int i, char c)
    {
	while (this.src.length < i)
	    realloc();
	
	this.src[i] = c;

	if (len <= i)
	    len = i;

	return this;
    }

    public char charAt(int i)
    {
	return src[i];
    }

    public int length()
    {
	return this.len;
    }

    public void length(int that_len)
    {
	while (this.src.length < that_len) 
	    realloc();

	this.len = that_len;
    }

    public KBuffer copy(int capacity)
    {
	try {

	    KBuffer clone = (KBuffer)super.clone();
	    char[] dst = new char[len + capacity];
	    System.arraycopy(src, 0, dst, 0, len);
	    clone.src = dst;
	    return clone;

	} catch (CloneNotSupportedException cnsex) {
	    cnsex.printStackTrace();
	    throw new InternalError();
	}
    }

    public void realloc()
    {
	if (DEBUG)
	    System.out.println("realloc to "+(src.length * 2));

	char[] dst = new char[src.length * 2];
	System.arraycopy(src, 0, dst, 0, len);
	this.src = dst;
    }

    public String toString()
    {
	return new String(this.src, 0, this.len);
    }

    public void toWriter(Writer out) throws IOException
    {
	synchronized (out) {
	    out.write(this.src, 0, this.len);
	}
    }

    public Object clone() throws CloneNotSupportedException
    {
	KBuffer clone = (KBuffer)super.clone();

	char[] dst = new char[len];
	System.arraycopy(src, 0, dst, 0, len);
	clone.src = dst;

	return clone;
    }

    public char[] getCharArray()
    {
	return src;
    }

    private int len;
    private char[] src;
}


