/**
 * $Id: JDK13ProxyClass.java,v 1.1 2001/07/27 02:07:00 pcj Exp $
 *
 * Copyright (C) 2001 Paul Cody Johnston - pcj@inxar.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA.
 **/
package org.inxar.hotswap;

import java.io.File;
import java.io.IOException;
import java.util.Stack;
import java.util.Iterator;
import java.util.List;
import java.util.LinkedList;
import java.util.ArrayList;
import java.lang.reflect.Method;
import java.lang.reflect.Constructor;

/**
 * Concrete <code>ProxyClass</code> implementation that supports
 * dynamic proxy classes.
 *
 * @since 0.8.2
 **/
class JDK13ProxyClass extends ProxyClass
{
    // ================================
    // Constructors 
    // ================================
    
    JDK13ProxyClass(ProxyClassLoader loader, String className)
    {
	super(loader, className);
	if (newProxyInstance == null)
	    throw new ProxyRuntimeException
		("Cannot instantiate: cannot find java.lang.reflect.Proxy.  "+
		 "Are you sure you are in a JDK1.3 environment?");

	this.interfaces = getProxyInterfaces(getClassInstance().getInterfaces());
	if (debug())
	    initDebug();
    }

    JDK13ProxyClass(ProxyClassLoader loader, String className, Class[] interfaces)
    {
	super(loader, className);

	if (newProxyInstance == null)
	    throw new ProxyRuntimeException
		("Cannot instantiate: cannot find java.lang.reflect.Proxy.  "+
		 "Are you sure you are in a JDK1.3 environment?");

	this.interfaces = getProxyInterfaces(interfaces);

	if (debug())
	    initDebug();
    }

    protected void initDebug()
    {
	log.debug("<init> New JDK1.3 ProxyClass created for "+className+":");
	for (int i = 0; i < interfaces.length; i++)
	    log.debug("       - implements "+interfaces[i].getName());
    }

    // ================================
    // ProxyClass Methods 
    // ================================

    public boolean isJDK13()
    {
	return true;
    }

    public Proxy newInstance()
    {
	return newJDK13Proxy(new KProxy(this), null);
    }

    public Proxy newInstance(Object[] args)
    {
	return newJDK13Proxy(new KProxy(this, args), null);
    }

    public Proxy newInstance(Class[] params, Object[] args)
    {
	return newJDK13Proxy(new KProxy(this, params, args), null);
    }
    
    public Proxy newInstanceH(Object h)
    {
	if (h == null)
	    throw new NullPointerException
		("A non-null ProxyInvocationHandler argument is required.");

	return newJDK13Proxy(new KProxy(this), h);
    }

    public Proxy newInstanceH(Object h, Object[] args)
    {
	if (h == null)
	    throw new NullPointerException
		("A non-null ProxyInvocationHandler argument is required.");

	return newJDK13Proxy(new KProxy(this, args), h);
    }

    public Proxy newInstanceH(Object h, Class[] params, Object[] args)
    {
	if (h == null)
	    throw new NullPointerException
		("A non-null ProxyInvocationHandler argument is required.");

	return newJDK13Proxy(new KProxy(this, params, args), h);
    }
    
    // ================================
    // Other Methods 
    // ================================

    protected void initDefaultInvocationHandler() 
    {
	List list = ProxyProperties.getProperties("invocationhandler.classname", 
					      getProxyClassLoader().getProperties());

	if (list != null && list.size() > 0) {
	    Iterator i = list.iterator();
	    boolean ok = false;
	    while (i.hasNext()) {
		ok = initDefaultInvocationHandler((String)i.next());
		if (ok) {
		    if (debug()) 
			log.debug("<initH> ProxyInvocationHandler is "+dh.getName());
		    break;
		}
	    }
	    if ( (!ok) && warn() ) 
		log.warn("<initH> Unable to use any of the listed "+
			 "ProxyInvocationHandler implementations ("+list+
			 ").  Falling back to default (ProxyInvocationHandler).");
	}

	// If one was not specified, use the default.
	if (dh == null)
	    dh = ProxyInvocationHandler.class;
    }

    protected boolean initDefaultInvocationHandler(String className)
    {
	try {
	    dh = Class.forName(className);
	    return true;
	} catch (Exception ex) {
	    return false;
	} 
    }

    protected Proxy newJDK13Proxy(Proxy proxy, Object h)
    {
	try {

	    // Instantiate a new default invocation handler if one has
	    // not been provided.
	    if (h == null) {
		if (dh == null)
		    initDefaultInvocationHandler();
		h = dh.newInstance();
	    }

	    // Call ProxyInvocationHandler.setProxy(proxy).
	    try {
		setProxy.invoke(h, new Object[]{ proxy });
	    } catch (IllegalArgumentException iaex) {
		throw new ProxyRuntimeException
		    ("InvocationHandler argument MUST "+
		     "extend ProxyInvocationHandler", iaex);
	    }

	    if (debug()) {
		log.debug("<newJDK13Proxy> Reflecting a new "+cls.getName()+" proxy to be serviced by ");
		log.debug(" an InvocationHandler of class "+h.getClass().getName()+"...");
	    }

	    // Finally, create a new hotswap Proxy using java proxy
	    // reflection.
	    return (Proxy)newProxyInstance.invoke(null, new Object[] { 
		getParentClassLoader(), 
		interfaces, 
		h 
	    });

	} catch (Exception ex) {
	    throw new ProxyRuntimeException(ex);
	}
    }

    protected static Class[] getProxyInterfaces(Class[] a)
    {
	if (a == null || a.length == 0) {
	    return new Class[]{ Proxy.class };
	} else {
	    // Make a new array such that we *also* implement Proxy
	    Class[] dst = new Class[a.length + 1];
	    System.arraycopy(a, 0, dst, 0, a.length);
	    dst[a.length] = Proxy.class;
	    return dst;
	}
    }

    // ================================
    // Fields Methods 
    // ================================

    /* interfaces including Proxy. */
    protected Class[] interfaces;

    // ================================
    // Class Fields and Methods 
    // ================================

    // org.inxar.hotswap.ProxyInvocationHandler
    static Class dh = null;

    // org.inxar.hotswap.ProxyInvocationHandler.setProxy
    static Method setProxy = null;

    // java.lang.reflect.Proxy.newProxyInstance
    static Method newProxyInstance = null; 

    static {
	try {

	    setProxy = Class.forName("org.inxar.hotswap.ProxyInvocationHandler")
		.getMethod("setProxy", new Class[]{ Proxy.class });

	    Class h_class = 
		Class.forName("java.lang.reflect.InvocationHandler");

	    newProxyInstance = Class.forName("java.lang.reflect.Proxy")
		.getMethod("newProxyInstance", new Class[] { 
		    ClassLoader.class, 
		    Class[].class, 
		    h_class
		});

	} catch (Exception ex) {
	    ex.printStackTrace();
	} 
    }
}

