/**
 * $Id: KProxy.java,v 1.7 2001/07/27 00:32:15 pcj Exp $
 *
 * Copyright (C) 2001 Paul Cody Johnston - pcj@inxar.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA.
 **/
package org.inxar.hotswap;

import java.util.List;
import java.util.LinkedList;
import java.util.Iterator;
import java.lang.reflect.Constructor;

/**
 * Concrete <code>Proxy</code> implementation.
 *
 * @since 0.8.2
**/
class KProxy implements Proxy 
{
    // ================================
    // Constructors 
    // ================================

    KProxy(ProxyClass cls, Class[] params, Object[] args)
    {
	this.cls = cls;

	if (params == null && args != null) {
	    params = new Class[args.length];
	    for (int i = 0; i < args.length; i++)
		params[i] = args[i].getClass();
	}

	if (cls.debug()) {
	    cls.log.debug("<KProxy.init> A new Proxy instance of "+cls.getName()+" has been created.");
	}

	hotswap_setConstructorParameters(params);
	hotswap_setConstructorArguments(args);
    }

    KProxy(ProxyClass cls, Object[] args)
    {
	this(cls, null, args);
    }

    KProxy(ProxyClass cls)
    {
	this(cls, null, null);
    }

    // ================================
    // TEHC Methods 
    // ================================

    public String toString()
    {
	return "(Proxy of class "+cls.getClassInstance().getName()+": "+obj+")";
    }

    public int hashCode()
    {
	return super.hashCode() ^ hash(params) ^ hash(args);
    }

    protected int hash(Object[] o)
    {
	int hash = 0;
	if (o != null)
	    for (int i = 0; i < o.length; i++) {
		if (o[i] != null) {
		    hash += o[i].hashCode();
		    hash ^= hash;
		}
	    }
	return hash;
    }

    public boolean equals(Object other)
    {
	if (other == this)
	    return true;

	if (!(other instanceof KProxy))
	    return false;

	KProxy that = (KProxy)other;

	// Compare class equality, parameter types, and argument types.
	return (this.cls.getClassInstance().equals(that.cls.getClassInstance())) &&
	    (equal(this.params, that.params)) &&
	    (equal(this.args, that.args));
    }

    protected boolean equal(Object[] _this, Object[] _that)
    {
	if (_this != null) {
	    if (_that != null) {
		if (_this.length != _that.length) 
		    return false;
		for (int i = 0; i < _this.length; i++) 
		    if (_this[i] == null) {
			if (_that[i] != null) 
			    return false;
		    } else if (!_this[i].equals(_that[i]))
			return false;
	    } else if (_this.length != 0)
		return false;
	} else if (_that != null && _that.length != 0)
	    return false;

	return true;
    }

    // ================================
    // Proxy Methods 
    // ================================

    public ProxyClass hotswap_getProxyClass()
    {
	return cls;
    }

    /* there may be deadlock conditions around this code */
    public Object hotswap() 
    {
	// If the object is null, this is the first time the object
	// has been constructed (or the arguments have been updated).
	// You way ask "What is the point of doing this now?"
	// Essentially because the enqueue operation wants a non-null
	// object. Also make new if the object has gone out of date.
	if (obj == null || !(obj.getClass().equals(cls.getClassInstance())) )
	    obj = hotswap_newInstance();

	hotswap_enqueue();
	cls.hotswap();

	return hotswap_getInstance();
    }

    public synchronized Object hotswap_getInstance() 
    {
	// If the object is null, this is the first time the object
	// has been constructed or the arguments have been updated.
	if (obj == null)
	    obj = hotswap_newInstance();

	return obj;
    }

    public synchronized int hotswap_getStatus()
    {
	return cls.getStatus();
    }

    public Object[] hotswap_getConstructorArguments()
    {
	return args;
    }

    public Class[] hotswap_getConstructorParameters()
    {
	return params;
    }

    public synchronized void hotswap_setConstructorArguments(Object[] args)
    {
	this.args = args;
	this.obj = null;	
    }

    public synchronized void hotswap_setConstructorParameters(Class[] params)
    {
	this.params = params;
	this.obj = null;
    }

    public synchronized void hotswap_addObjectSwapListener(ProxyEventListener l)
    {
	if (listeners == null)
	    listeners = new LinkedList();
	listeners.add(l);
    }

    public synchronized void hotswap_removeObjectSwapListener(ProxyEventListener l)
    {
	if (listeners != null) 
	    listeners.remove(l);
    }

    public synchronized void hotswap_enqueue()
    {
	if (!cls.isEnqueued(this))
	    cls.enqueue(this);
    }

    public synchronized void hotswap_dequeue()
    {
	if (cls.isEnqueued(this))
	    cls.dequeue(this);
    }

    public synchronized boolean hotswap_isEnqueued()
    {
	return cls.isEnqueued(this);
    }

    public synchronized void hotswap_release()
    {
	// Release from the proxy class
	cls.dequeue(this);

	if (tmp instanceof ProxyObject) {
	    if (cls.debug()) {
		cls.log.debug("<hotswap_release> Releasing old ProxyObject " +
			      System.identityHashCode(obj) + "...");
	    }
	    ((ProxyObject)tmp).hotswap_onRelease();
	}
	tmp = null;
    }

    // ================================
    // Package Methods 
    // ================================

    synchronized boolean hotswap_prepare(Class newClass)
    {
	try {

	    tmp = newClass.getDeclaredConstructor(params)
		.newInstance(args);

	} catch (Exception ex) {
	    ex.printStackTrace();
	    return false;
	}

	// Prepare this instance.
	if (tmp instanceof ProxyObject && obj instanceof ProxyObject) {
	    if (cls.debug()) {
		cls.log.debug("<hotswap_prepare> Preparing ProxyObject " +
			      System.identityHashCode(tmp) + "...");
	    }
	    return ((ProxyObject)tmp).hotswap_onPrepare((ProxyObject)obj);
	} else
	    return true;
    }

    synchronized void hotswap_commit()
    {
	if (tmp instanceof ProxyObject) {
	    if (cls.debug()) {
		cls.log.debug("<hotswap_commit> Committing ProxyObject " + 
			      System.identityHashCode(tmp) + "...");
	    }
	    ((ProxyObject)tmp).hotswap_onCommit();
	} 

	if (obj instanceof ProxyObject) {
	    if (cls.debug()) {
		cls.log.debug("<hotswap_commit> Releasing old ProxyObject " + 
			      System.identityHashCode(obj) + "...");
	    }
	    ((ProxyObject)obj).hotswap_onRelease();
	}

	// Swap the old instance with the new.
	Object old = obj;
	obj = tmp;
	tmp = old;

	fire( new ProxyObjectSwapEvent(this, old, obj) );
    }

    synchronized void hotswap_rollback()
    {
	if (tmp instanceof ProxyObject) {
	    if (cls.debug()) {
		cls.log.debug("<hotswap_rollback> Rolling back ProxyObject " + 
			      System.identityHashCode(obj) + "...");
	    }
	    ((ProxyObject)tmp).hotswap_onRollback();
	}
	tmp = null;
    }

    // ================================
    // Protected Methods 
    // ================================

    protected synchronized Object hotswap_newInstance()
    {
	try {
	    
	    if (cls.debug()) {
		cls.log.debug("<hotswap_newInstance> Instantiating new " + 
			      cls.getName() + " object...");
	    }

  	    return cls.getClassInstance()
  		.getDeclaredConstructor(params)
  		.newInstance(args);

	} catch (Exception ex) {
	    throw new ProxyRuntimeException(ex);
	} 

    }

    protected void fire(ProxyObjectSwapEvent evt)
    {
	if (listeners != null && listeners.size() > 0) {
	    Iterator i = listeners.iterator();
	    while (i.hasNext()) 
		((ProxyEventListener)i.next()).notify(evt);
	}
    }

    // ================================
    // Other Methods 
    // ================================

    protected Class[] params;	
    protected Object[] args;	
    protected ProxyClass cls;	
    protected List listeners;	
    protected Object obj;	
    protected Object tmp;	
}
