/**
 * $Id: PriorityQueue.java,v 1.1 2001/07/27 02:07:00 pcj Exp $
 *
 * Copyright (C) 2001 Paul Cody Johnston - pcj@inxar.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA.
 **/
package org.inxar.hotswap;

/**
 * The <code>PriorityQueue</code> is a heap-based priority queue used
 * to schedule recurrent notification events.
 *
 * @since 0.8.6
**/
class PriorityQueue
{
    protected static final boolean DEBUG = true;
    
    // ================================
    // Constructors
    // ================================

    public PriorityQueue(ProxyClassLoader loader)
    {
	this.loader = loader;
	this.a = new Node[5];
	this.len = 0;
	this.notifier = new Notifier();
	this.notifier.start();
    }

    // ================================
    // TEHC Methods
    // ================================

    public String toString()
    {
	StringBuffer b = new StringBuffer();
	b.append('[');
	for (int i = 0; i < len; i++) {
	    if (i > 0)
		b.append(',');
	    b.append(a[i]);
	}
	b.append(']');
	return b.toString();
    }

    // ================================
    // Scheduling Methods
    // ================================

    //public void schedule(AlarmEventListener listener, Object callback, long period)
    public void schedule(Listener listener, Object callback, long period)
    {
	Node node = this.new Node(listener, callback, period);
	insert(node);
    }

    public void ping()
    {
	System.out.println("Notifier Thread Dump:");
	System.out.println("Is notifier null? "+notifier);
	System.out.println("Is clock null? "+notifier.clock);
	System.out.println("Is clock alive? "+notifier.clock.isAlive());
	notifier.clock.dumpStack();
    }

    // ================================
    // Heapsort Methods
    // ================================

    protected void heapify(int i)
    {
	int l = left(i);
	int r = right(i);
	int min = (l < len && a[l].key < a[i].key) ? l : i;

	if (r < len && a[r].key < a[min].key)
	    min = r;

	if (min != i) {
	    Node tmp = a[i]; a[i] = a[min]; a[min] = tmp;
	    heapify(min);
	}
    }

    protected synchronized void buildHeap()
    {
	for (int i = (len >> 1) - 1; i >= 0; --i) 
	    heapify(i);
    }

    protected synchronized void heapSort()
    {
	int oldlen = len;

	buildHeap();

	Node tmp;
	for (int i = len - 1; i > 0; --i) {
	    tmp = a[i]; a[i] = a[0]; a[0] = tmp;
	    --len;
	    heapify(0);
	}

	len = oldlen;
    }

    protected synchronized void insert(Node n)
    {
	n.key = System.currentTimeMillis() + n.period;

	if (len == a.length)
	    realloc();

	int i = len++;
	int p;
	while (i >= 0) {
	    p = parent(i);

	    if (p < 0) {
		//System.out.println("insert break: parent of "+i+" is "+p);
		break;
	    }

	    if (a[p].key < n.key) {
		//System.out.println("insert break: parent key ("+a[p].key+
		//		   ") less than node key ("+n.key+")");
		break;
	    }
	    
	    a[i] = a[p];
	    i = p;
	}

	a[i] = n;

	//System.out.println("Node "+n+" inserted at position "+i);
	//System.out.println(this);

	if (i == 0) { 
	    //System.out.println("interrupting now");
	    notifier.interrupt();
	}
    }

    protected synchronized Node extractMinNode()
    {
	if (len == 0)
	    throw new IndexOutOfBoundsException();

	Node min = a[0];
	a[0] = a[--len];
	heapify(0);
	return min;
    }

    protected int parent(int i)
    {
	if (i == 1)
	    return 0;

	return (i >> 1) - 1;
    }

    protected int left(int i)
    {
	return (i << 1) | 1;
    }

    protected int right(int i)
    {
	return (i << 1) + 2;
    }

    protected void realloc()
    {
	Node[] d = new Node[a.length * 2];
	System.arraycopy(a, 0, d, 0, a.length);
	a = d;
    }

    public int size()
    {
	return len;
    }

    public void stop()
    {
	try {
	    notifier.stop();
	    notifier.interrupt();
	} catch (Exception ex) {
	} 
    }

    // ================================
    // Debugging Methods 
    // ================================

    protected boolean debug()
    {
	if (lp == 0)
	    return false;

	if (lp < 0) 
	    initChannel();
	
	return (lp & ConsoleLog.PRI_DEBUG) != 0;
    }

    protected void initChannel()
    {
	if (loader.hasLog()) {
	    log = loader.getLog().newChannel("que", this);
	    lp = ConsoleLog.encode(log);
	} else {
	    lp = 0;
	}
    }

    // ================================
    // Fields
    // ================================

    protected int len;
    protected Node[] a;
    protected Notifier notifier;

    protected ProxyClassLoader loader;
    protected ProxyLog.Channel log;	
    protected int lp = -1; 

    // ================================
    // Interfaces
    // ================================

    /**
     * Clients need to implement this interface.
    **/
    interface Listener
    {
	void queue_notify(Object callback);
    }

    // ================================
    // Inner Classes
    // ================================

    class Notifier implements Runnable
    {
	Notifier()
	{
	}
	
	public void run()
	{
	    // FYI: When a thread is interrupted, it goes back into
	    // the run method directly?  Bizarre!

	    Thread thisThread = Thread.currentThread();

	    while(thisThread == clock) {

		try {

		    long delay = Long.MAX_VALUE;

		    if (size() > 0) {
			do {
			    //System.out.println(AlarmQueue.this);
			    long now = System.currentTimeMillis();
			    long then = a[0].key;
			    delay = then - now;

			    //System.out.println("Delay is "+delay+" ms.");
			    // If (then - now) is positive, then it
			    // means that we still have some time to
			    // wait before doing the notification
			    // (then is a time greater than now). 
			    if (delay > 20)
				break;
			    
			    // If we do not break, it means that the
			    // current node needs to be notified.
			    Node n = extractMinNode();
			    n.notifyListener();
			    insert(n);

			    //ping();

			} while (true);
		    }

		    if (debug())
			log.debug("Sleeping for "+delay+" ms...");

		    clock.sleep(delay);

		} catch (InterruptedException iex) {
		}

	    }

	    if (debug())
		log.debug("Exiting run()...  Goodbye.");
	}

	public void stop()
	{
	    clock = null;
	}
	
	public void start()
	{
	    if (clock == null) {
		clock = new Thread(this);
		clock.setDaemon(true);
		clock.setName("Notifier clock");
		clock.start();
	    }
	}

	public void interrupt()
	{
	    clock.interrupt();
	}
	
	Thread clock;
    }

    class Node 
    {
	Node(PriorityQueue.Listener listener, Object callback, long period) 
	{
	    this.listener = listener;
	    this.callback = callback;
	    this.period = period;
	}
	
	public String toString()
	{
	    StringBuffer b = new StringBuffer();
	    b.append('(');
	    b.append(callback);
	    b.append(',');
	    long time = key - System.currentTimeMillis();
	    b.append(Math.abs(time));
	    b.append(time > 0 ? " early" : " late");
	    b.append(')');
	    return b.toString();
	}
	
	public void notifyListener()
	{
	    if (debug())
		log.debug(listener + " notified within "+
			  (System.currentTimeMillis() - key) + " ms of target time.");
	    
	    listener.queue_notify(callback);
	}
	
	long key;		// next scheduled notification time
	long period;		// time between notifications 
	Object callback;	// callback object
	PriorityQueue.Listener listener; // listener object
    }
}



