/**
 * $Id: ProxyClass.java,v 1.10 2001/07/27 02:07:00 pcj Exp $
 *
 * Copyright (C) 2001 Paul Cody Johnston - pcj@inxar.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA.
 **/
package org.inxar.hotswap;

import java.util.Date;
import java.util.Stack;
import java.util.Iterator;
import java.util.List;
import java.util.LinkedList;
import java.util.ArrayList;

/**
 * The <code>ProxyClass</code> is as a factory for <code>Proxy</code>
 * instances and holds a reference to a <code>Class</code> whose
 * implementation may change during the life of an application.
 *
 * @since 0.8.2
**/
public abstract class ProxyClass
{
    private static final boolean LOG = true;

    // ================================
    // Constructors 
    // ================================

    ProxyClass(ProxyClassLoader loader, String className)
    {
	this.loader = loader;
	this.className = className;
	this.proxies = new Stack();
    }

    // ================================
    // TEHC Methods 
    // ================================

    // ================================
    // Abstract Methods 
    // ================================

    /**
     * JDK1.2 or JDK1.3 Factory method; Creates a new
     * <code>Proxy</code> for this <code>ProxyClass</code> having no
     * constructor arguments (though arguments can be set later on the
     * <code>Proxy</code> itself).
    **/
    public abstract Proxy newInstance();

    /**
     * JDK1.2 or JDK1.3 Factory method; Creates a new
     * <code>Proxy</code> having the given constructor arguments.  The
     * parameters are gathered by fetching the <code>Class</code> for
     * each argument.
    **/
    public abstract Proxy newInstance(Object[] args);

    /**
     * JDK1.2 or JDK1.3 Factory method; Creates a new
     * <code>Proxy</code> having the given constructor parameters and
     * arguments.
    **/
    public abstract Proxy newInstance(Class[] params, Object[] args);

    /**
     * JDK1.3 Factory method; Creates a new <code>Proxy</code> for
     * this <code>ProxyClass</code> having no constructor arguments
     * using the given <code>ProxyInvocationHandler</code> (H) instance.
     *
     * @exception <code>UnsupportedOperationException</code> if this
     * is not a JDK1.3 proxy class.
    **/
    public abstract Proxy newInstanceH(Object h);

    /**
     * JDK1.3 Factory method; Creates a new <code>Proxy</code> having
     * the given constructor arguments using the given
     * <code>ProxyInvocationHandler</code> (H) instance.  The
     * parameters are gathered by fetching the <code>Class</code> for
     * each argument.
     *
     * @exception <code>UnsupportedOperationException</code> if this
     * is not a JDK1.3 proxy class.
    **/
    public abstract Proxy newInstanceH(Object h, Object[] args);

    /**
     * JDK1.3 Factory method; Creates a new <code>Proxy</code> having
     * the given constructor parameters and arguments using the given
     * <code>ProxyInvocationHandler</code> (H) instance. This
     * invocation handler will be passed to the
     * <code>java.lang.reflect.Proxy.newProxyInstance</code> method
     * and MUST be assignable to
     * <code>org.inxar.hotswap.ProxyInvocationHandler</code>.  <P>The
     * parameter type is given here as <code>Object</code> rather than
     * <code>ProxyInvocationHandler</code> in order to avoid
     * classloading of <code>JDK1.3</code> dependent classes in a
     * <code>JDK1.2</code> environment.
     *
     * @exception <code>UnsupportedOperationException</code> if this
     * is not a JDK1.3 proxy class.
    **/
    public abstract Proxy newInstanceH(Object h, Class[] params, Object[] args);

    // ================================
    // Public Methods 
    // ================================

    /**
     * Returns the parent <code>ProxyClassLoader</code> that created
     * this <code>ProxyClass</code>.
     **/
    public ProxyClassLoader getProxyClassLoader()
    {
	return loader;
    }

    /**
     * Returns the parent <code>ClassLoader</code>.  If none has been
     * explicitly set, the return values is
     * <code>getProxyClassLoader().getDefaultParentClassLoader()</code>.
     **/
    public ClassLoader getParentClassLoader()
    {
	return parent != null ? parent
	    : loader.getDefaultParentClassLoader();
    }

    /**
     * Sets the parent <code>ClassLoader</code> to be used by this
     * class.  This <code>ClassLoader</code> is used as the
     **/
    public void setParentClassLoader(ClassLoader parent)
    {
	this.parent = parent;
    }

    /**
     * Returns the <code>ProxyCompiler</code> used by this class.
     **/
    public ProxyCompiler getProxyCompiler()
    {
	return compiler != null 
	    ? compiler
	    : loader.getDefaultProxyCompiler();
    }

    /**
     * Sets the <code>ProxyCompiler</code> to be used by this class.
     **/
    public void setProxyCompiler(ProxyCompiler compiler)
    {
	this.compiler = compiler;
    }

    /**
     * Returns <code>true</code> if this method is a
     * <code>JDK1.3</code> proxy class.
    **/
    public boolean isJDK13()
    {
	return false;
    }

    /**
     * Returns the timestamp of the last successful classload or
     * <code>0L</code> if no class has been loaded.
    **/
    public synchronized long getLastClassload()
    {
	return loadStamp;
    }

    /**
     * Returns the Sourcefile <code>ProxyResource</code> for this
     * <code>ProxyClass</code>.
    **/
    public ProxyResource getSourcefile()
    {
	return getProxyCompiler().getSourcefile(this);
    }

    /**
     * Returns the Sourcefile <code>ProxyResource</code> for this
     * <code>ProxyClass</code>.
    **/
    public ProxyResource getClassfile()
    {
	return getProxyCompiler().getClassfile(this);
    }

    /**
     * Returns the name of the class being proxied; equivalent to
     * <code>ProxyClass.getClassInstance().getName()</code>.
    **/
    public String getName()
    {
	return className;
    }
    
    /**
     * Returns the status of this <code>ProxyClass</code> as
     * determined by the <code>ProxyCompiler.getStatus</code> method.
    **/
    public synchronized int getStatus()
    {
	return getProxyCompiler().getStatus(this);
    }

    /**
     * Returns the <code>Class</code> instance that this
     * <code>ProxyClass</code> contains.
    **/
    public Class getClassInstance() 
    {
	if (cls == null) {
	    // See what to do
	    switch (getStatus()) {
		
	    case ProxyCompiler.STATUS_UNKNOWN:
		/* if unknown what to do bail */
		throw new ProxyRuntimeException
		    (new ClassNotFoundException("Cound not stat "+className));
		
	    case ProxyCompiler.STATUS_COMPILE:
		// Do we want to check the return value?  If it fails,
		// we still want to try to load a class.
		ProxyCompileEvent e = getProxyCompiler().compile(this);
		if (hasCompileListeners())
		    fire(e);

		/* fall-through */

	    case ProxyCompiler.STATUS_LOAD:
	    case ProxyCompiler.STATUS_CURRENT:
		cls = newClassInstance();
		break;

	    case ProxyCompiler.STATUS_ERROR:
		throw new ProxyRuntimeException
		    ("Unable to load class "+className+
		     "! (are you sure this is the right name?) "+
		     "Please check log.debug for details.");

	    default:
		throw new InternalError(); /* bad coding */
	    }

	}
	return cls;
    }

    /**
     * Adds the given listener of <code>ProxyHotSwapEvent</code>
     * notifications.
    **/
    public synchronized void addHotSwapListener(ProxyEventListener l)
    {
	if (hotSwapListeners == null)
	    hotSwapListeners = new LinkedList();
	hotSwapListeners.add(l);
    }

    /**
     * Removes the given listener of <code>ProxyHotSwapEvent</code>
     * notifications.
    **/
    public synchronized void removeHotSwapListener(ProxyEventListener l)
    {
	if (hotSwapListeners != null) 
	    hotSwapListeners.remove(l);
    }

    /**
     * Adds the given listener of <code>ProxyCompileEvent</code>
     * notifications.
    **/
    public synchronized void addCompileListener(ProxyEventListener l)
    {
	if (compileListeners == null)
	    compileListeners = new LinkedList();
	compileListeners.add(l);
    }

    /**
     * Removes the given listener of <code>ProxyCompileEvent</code>
     * notifications.
    **/
    public synchronized void removeCompileListener(ProxyEventListener l)
    {
	if (compileListeners != null) 
	    compileListeners.remove(l);
    }

    /**
     * Adds the given listener of <code>ProxyClassSwapEvent</code>
     * notifications.
    **/
    public synchronized void addClassSwapListener(ProxyEventListener l)
    {
	if (classSwapListeners == null)
	    classSwapListeners = new LinkedList();
	classSwapListeners.add(l);
    }

    /**
     * Removes the given listener of <code>ProxyClassSwapEvent</code>
     * notifications.
    **/
    public synchronized void removeClassSwapListener(ProxyEventListener l)
    {
	if (classSwapListeners != null) 
	    classSwapListeners.remove(l);
    }

    /**
     * Executes a hotswap transaction.  See the package description
     * for more details.
    **/
    public boolean hotswap()
    {
	synchronized (this) {
	    // First of all, if there are no proxies that will be involved
	    // in the hotswap, then dont bother. Only hotswap if it will
	    // make a difference to the end user.
	    if (!hasProxies())
		return false;
	    
	    // Disallow recursive hotswaps.
	    if (isHotswapping) {
		if (warn()) 
		    log.warn("<hotswap> Recursive HotSwapping detected! This is not allowed.");
		return false;
	    } 
	    
	    isHotswapping = true;
	}

	boolean result = false;

	try {

	    /* do not want the notification section to be
	       synchronized: if it were, we'd get a deadlock when
	       another thread tries to access this ProxyClass object
	       on the event. */

	    // fire off a ProxyHotSwapEvent if needed.
	    if (hasHotSwapListeners()) 
		fire( new ProxyHotSwapEvent(this) ); 

	    result = hotswapInternal();
	    
	    /* NOTE: The KProxy typically registers itself here before
               calling cls.hotswap().  If an exception is thrown here,
               the stack will still contain that proxy for later
               calls.  Is this OK?  */

	} finally {
	    isHotswapping = false;
	}

	return result;
    }

    protected synchronized boolean hotswapInternal()
    {
	// See what to do
	switch (getStatus()) {

	case ProxyCompiler.STATUS_UNKNOWN:
	    /* if unknown what to do return false by definition */
	    return false;

	case ProxyCompiler.STATUS_CURRENT:
	    /* if nothing needs to be done return true by definition */
	    return true;

	case ProxyCompiler.STATUS_LOAD:
	    /* Break switch and continue in this method */
	    break;

	case ProxyCompiler.STATUS_COMPILE:
	    /* try to compile */
	    ProxyCompileEvent e = getProxyCompiler().compile(this);
	    if (hasCompileListeners())
		fire(e);
	    if (e.getReturnCode() == ProxyCompiler.COMPILE_SUCCESS)
		/* If the compilation succeeds, break switch and
		   continue in this method */
		break;
	    else
		/* If the compilation fails return false
		   by definition */
		return false;

	case ProxyCompiler.STATUS_ERROR:
	    throw new ProxyRuntimeException("Unable to load class "+className+
					    "! (are you sure this is the correct name and location?) "+
					    "Please check log.debug for details.");

	default:
	    throw new InternalError(); /* bad coding */
	}

	// Ok, next step is to reload the class.
	Class oldClass = this.cls;
	Class newClass = null;

	try {
	    
	    // If the load succeeds, the classloader will cache the
	    // object automatically.
	    newClass = newClassInstance();

	} catch (Exception ex) {
	    
	    /* Remove all proxies, the hotswap died. */
	    proxies.clear();
	    return false;
	    
	}

	// Create a new list that will contain all successfully
	// prepared proxies.
	List prepared = new ArrayList(proxies.size());
	Iterator i;
	KProxy p;
	boolean commit = true;
	try {

	    // First prepare all proxies.
	    while (!proxies.isEmpty()) {
		p = (KProxy)proxies.pop();
		if (p.hotswap_prepare(newClass)) 
		    prepared.add(p);
		else
		    commit = false;
	    }

	} catch (Exception ex) {
	    /* might be nice to know the error */
	    System.out.println("Exception caught during hotswap transaction preparation:");
	    ex.printStackTrace();
	    commit = false;
	}
	
	if (commit) {

	    // Do the commit
	    i = prepared.iterator();
	    while (i.hasNext())
		((KProxy)i.next()).hotswap_commit();

	    // The transaction has succeeded.  Drop the old class and
	    // replace it with the new. Also update classload
	    // timestamp.
	    this.loadStamp = System.currentTimeMillis();
	    this.cls = newClass;

	    if (debug())
		log.debug("<hotswap> Success! Last classload now dated "+new Date(loadStamp).toString());

	    // Finally, notify the classSwapListeners.
	    if (hasClassSwapListeners()) 
		fire( new ProxyClassSwapEvent(this, oldClass, newClass) );

	} else {

	    if (debug())
		log.debug("<hotswap> Commit failed, rolling back...");

	    // Rollback any proxies that successfully prepared.
	    if (prepared.size() > 0) {
		i = prepared.iterator();
		while (i.hasNext())
		    ((KProxy)i.next()).hotswap_rollback();
	    }

	}

	return commit;
    }

    // ================================
    // Package Methods 
    // ================================

    synchronized boolean hasHotSwapListeners()
    {
	return hotSwapListeners != null && hotSwapListeners.size() > 0;
    }

    synchronized boolean hasCompileListeners()
    {
	return compileListeners != null && compileListeners.size() > 0;
    }

    synchronized boolean hasClassSwapListeners()
    {
	return classSwapListeners != null && classSwapListeners.size() > 0;
    }

    synchronized boolean hasProxies()
    {
	return proxies.size() > 0;
    }

    synchronized void enqueue(KProxy proxy)
    {
	if (!proxies.contains(proxy))
	    proxies.add(proxy);
    }

    synchronized void dequeue(KProxy proxy)
    {
	proxies.remove(proxy);
    }

    synchronized boolean isEnqueued(KProxy proxy)
    {
	return proxies.contains(proxy);
    }

    // ================================
    // Protected Methods 
    // ================================

    /**
     * Returns a new intance of this <code>Class</code>.
    **/
    protected synchronized Class newClassInstance()
    {
	try {

	    if (debug()) 
		log.debug("<newClassInstance> Attempting classload of "+className+"...");
	    
	    return new ProxyClassClassLoader(this).loadClass(className);

	} catch (ClassNotFoundException ex) {
	    throw new ProxyRuntimeException(ex);
	} 
    }

    protected synchronized void fire(ProxyHotSwapEvent evt)
    {
	if (hasHotSwapListeners()) {
	    Iterator i = hotSwapListeners.iterator();
	    while (i.hasNext()) 
		((ProxyEventListener)i.next()).notify(evt);
	}
    }

    protected synchronized void fire(ProxyCompileEvent evt)
    {
	if (hasCompileListeners()) {
	    Iterator i = compileListeners.iterator();
	    while (i.hasNext()) 
		((ProxyEventListener)i.next()).notify(evt);
	}
    }

    protected synchronized void fire(ProxyClassSwapEvent evt)
    {
	if (hasClassSwapListeners()) {
	    Iterator i = classSwapListeners.iterator();
	    while (i.hasNext()) 
		((ProxyEventListener)i.next()).notify(evt);
	}
    }

    // ================================
    // Debugging Methods 
    // ================================

    protected boolean fatal()
    {
	if (lp == 0)
	    return false;

	if (lp < 0) 
	    initChannel();
	
	return (lp & ConsoleLog.PRI_FATAL) != 0;
    }

    protected boolean warn()
    {
	if (lp == 0)
	    return false;

	if (lp < 0) 
	    initChannel();
	
	return (lp & ConsoleLog.PRI_WARN) != 0;
    }

    protected boolean debug()
    {
	if (lp == 0)
	    return false;

	if (lp < 0) 
	    initChannel();
	
	return (lp & ConsoleLog.PRI_DEBUG) != 0;
    }

    protected void initChannel()
    {
	if (loader.hasLog()) {
	    log = loader.getLog().newChannel("cls", this);
	    lp = ConsoleLog.encode(log);
	} else {
	    lp = 0;
	}
    }

    // ================================
    // Fields 
    // ================================

    protected String className;	// the name of the class to proxy
    protected Class cls;	// the instance of the class to proxy
    protected Stack proxies;	// a list of enqueued proxies
    protected long loadStamp;	// the classload timestamp
    protected ProxyCompiler compiler; // parent compiler
    protected ClassLoader parent; // parent classloader
    protected ProxyClassLoader loader; // parent proxyclassloader
    protected boolean isHotswapping; // flag used to detect hotswap recursion
    protected List hotSwapListeners; // list of ProxyEventListeners
    protected List compileListeners; // list of ProxyEventListeners
    protected List classSwapListeners; // list of ProxyEventListeners

    // The log channel for this class
    protected ProxyLog.Channel log;	

    // The log priorities
    protected int lp = -1; 
}

/*

Better not to have to choose a compiler type because that gets
hardcoded into your application.  If the application is moved to a
differnt environment, you have to update the code.

 **/

//  	if (LOG) {
//  	    channel = compiler.getLog().newChannel("cls", this);
//  	    channel.debug(new StringBuffer()
//  		.append("ProxyClass initialized for class ")
//  		.append(className)
//  		.append(" with srcfile `")
//  		.append(sourceFile.toString())
//  		.append("' and classfile `")
//  		.append(classFile.toString())
//  		.append("'."));
//  	}
