/**
 * $Id: ProxyClassLoader.java,v 1.2 2001/07/27 02:07:00 pcj Exp $
 *
 * Copyright (C) 2001 Paul Cody Johnston - pcj@inxar.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA.
 **/
package org.inxar.hotswap;

import java.io.InputStream;
import java.io.FileInputStream;
import java.io.BufferedInputStream;
import java.io.IOException;

import java.net.URL;

import java.util.List;
import java.util.Iterator;
import java.util.Enumeration;
import java.util.Map;
import java.util.HashMap;
import java.util.Properties;
import java.util.Collection;
import java.util.Collections;

/**
 * The <code>ProxyClassLoader</code> acts as a factory of
 * <code>ProxyClass</code> instances.  It is the "top-level" object in
 * the library.  Note that the name of this class means "a thing that
 * loads <code>ProxyClass</code> instances" and not "a thing that is a
 * <code>ClassLoader</code>".  This class does not extend
 * <code>java.lang.ClassLoader</code>.
 *
 * <P>
 * <table border=0 cellpadding="0" cellspacing="0" bgcolor="#999999" width="100%">
 * <tr><td>
 * <table cellpadding="0" cellspacing="1"><tr><td><font face="helvetica" color="#eeeeee">
 * ProxyCompiler Usage
 * </font></td></tr></table>
 * </td></tr>
 * <tr><td>
 * <table border=0 cellpadding="7" cellspacing="1" width="100%"><tr bgcolor="#eeeeff" ><td>
 * <pre>
 * ProxyClassLoader loader = new ProxyClassLoader();
 *
 * ProxyClass = loader.loadJDK12("org.myname.myproject.MyClass");
 * ProxyClass = loader.load("org.myname.myproject.MyClass");
 * </td></tr></table>
 * </td></tr>
 * </table>
 * <P>
 *
 * @since 0.8.6
**/
public class ProxyClassLoader
{
    private static final boolean LOG = true;

    // ================================
    // Constructors 
    // ================================

    public ProxyClassLoader() throws ProxyException
    {
	InputStream in = null;

	try {

	    // See if they provided a -D option.
	    String origin = null;
	    String name = System.getProperty("org.inxar.hotswap.properties");

	    if (name != null) {
		in = new FileInputStream(name);
		origin = "file:"+name;

	    } else {
		name = "org/inxar/hotswap/hotswap.properties";
		URL url = ProxyClassLoader.class.getClassLoader().getResource(name); 
		if (url == null)
		    throw new ProxyException
			("Could not find properties file resource `"+name+"'");

		origin = url.toString();
		in = url.openStream();
	    }
		
	    initProperties(in);
	    initSystemProperties();
	    config.put("origin", origin);
	    init();

	    in.close();
	    in = null;

	} catch (IOException ioex) {
	    if (in != null)
		try { in.close(); } 
		catch (Exception ex) {} 
	    if (ioex instanceof ProxyException)
		throw (ProxyException)ioex;
	    else
		throw new ProxyException(ioex);
	} 
    }

    public ProxyClassLoader(String propertiesFileName) throws ProxyException
    {
	InputStream in = null;
	try {
	    
	    in = new FileInputStream(propertiesFileName);

	    initProperties(in);
	    initSystemProperties();
	    init();

	    in.close();
	    in = null;

	} catch (IOException ioex) {
	    if (in != null)
		try { in.close(); } 
		catch (Exception ex) {} 
	    if (ioex instanceof ProxyException)
		throw (ProxyException)ioex;
	    else
		throw new ProxyException(ioex);
	} 
    }

    // ================================
    // Initialization Methods
    // ================================

    /**
     * Called upon object instantiation.  The ProxyClassLoader expects
     * to initialize state from a Properties object from a file; this
     * method creates that Properties object .  
    **/
    protected void initProperties(InputStream in) throws IOException
    {
	config = new ProxyProperties();
	config.load(in);
    }

    protected void initSystemProperties() 
    {
	// Transfer in all the properties that were specified using
	// the -D option at startup.
	String prefix = "org.inxar.hotswap.";
	int len = prefix.length();

	Properties p = System.getProperties();
	Enumeration e = p.propertyNames();
	String k;
	while (e.hasMoreElements()) {
	    k = (String)e.nextElement();
	    if (k.startsWith(prefix)) 
		config.push(k.substring(len), p.getProperty(k));
	}
    }

    /**
     * Called after the Properties has been created. This object
     * contains either String or List<String> objects.
    **/
    protected void init() throws ProxyException
    {
	this.proxyClasses = new HashMap();

	initProxyLog();
	initParentClassLoader();
	initProxyCompiler();
    }

    /**
     * Setup the <code>ProxyLog</code>.
    **/
    protected void initProxyLog() throws ProxyException
    {
	List list = ProxyProperties.getProperties("log.classname", config);

	if (list != null && list.size() > 0) {
	    Iterator i = list.iterator();
	    boolean ok = false;
	    while (i.hasNext()) {
		ok = initProxyLog((String)i.next());
		if (ok) {
		    if (debug()) {
			log.debug("Hello, welcome to HotSwap version "+getVersion()+"!");
			log.debug("- homepage is at http://www.inxar.org/hotswap");
			log.debug("- latest version is at http://www.inxar.org/download");
			log.debug("- please report bugs and feedback to <hotswap@inxar.org>");
			log.debug("");
			log.debug("<init> java.version is "+System.getProperty("java.version"));
			log.debug("<init> java.class.path is "+System.getProperty("java.class.path"));
			log.debug("<init> Properties loaded from "+config.get("origin"));
			log.debug("<init> ProxyLog is "+master);
		    }
		    break;
		}
	    }
	    if (!ok)
		throw new ProxyException
		    ("Unable to instantiate ProxyLog from: "+list);
	}
    }

    protected boolean initProxyLog(String className)
    {
	try {
	    master = (ProxyLog)Class.forName(className).newInstance();
	    return master.initialize(config);
	} catch (Exception ex) {
	    return false;
	} 
    }

    /**
     * Called to setup the parent classloader.
    **/
    protected void initParentClassLoader() throws ProxyException
    {
	List list = ProxyProperties.getProperties("classloader.classname", config);

	if (list != null && list.size() > 0) {
	    Iterator i = list.iterator();
	    boolean ok = false;
	    while (i.hasNext()) {
		ok = initParentClassLoader((String)i.next());
		if (ok) {
		    if (debug())
			log.debug("<init> Default parent ClassLoader is "+parent); 
		    break;
		}
	    }
	    if (!ok)
		throw new ProxyException
		    ("Unable to instantiate parent ClassLoader from: "+list);
	} else 
	    parent = ProxyClassLoader.class.getClassLoader();
    }

    protected boolean initParentClassLoader(String className)
    {
	try {
	    parent = Class.forName(className).getClassLoader();
	    return true;
	} catch (Exception ex) {
	    return false;
	} 
    }

    protected void initProxyCompiler() throws ProxyException
    {
	List list = ProxyProperties.getProperties("compiler.classname", config);

	if (list != null && list.size() > 0) {
	    Iterator i = list.iterator();
	    boolean ok = false;
	    while (i.hasNext()) {
		ok = initProxyCompiler((String)i.next());
		if (ok) {
		    if (debug())
			log.debug("<init> Default ProxyCompiler is "+compiler); 
		    break;
		}
	    }
	    if (!ok)
		throw new ProxyException
		    ("Unable to instantiate ProxyCompiler from: "+list);
	} 
    }

    protected boolean initProxyCompiler(String className)
    {
	try {
	    compiler = (ProxyCompiler)Class.forName(className).newInstance();
	    compiler.setProxyClassLoader(this);
	    return compiler.initialize(config);
	} catch (Exception ex) {
	    return false;
	} 
    }

    // ================================
    // TEHC Methods
    // ================================

    // ================================
    // Public Methods 
    // ================================

    /**
     * Returns the version number for this release.
    **/
    public String getVersion()
    {
	return "0.8.6";
    }

    /**
     * Sets the <code>ProxyLog</code> for this
     * <code>ProxyClassLoader</code> and her daughter
     * <code>ProxyClass</code> instances.  Pass an instance that
     * implements the <code>ProxyLog</code> interface to hook HotSwap
     * into the logging system in your application.
    **/
    public synchronized void setLog(ProxyLog master)
    {
	this.master = master;
    }

    /**
     * Returns the <code>ProxyLog</code> or <code>null</code> if one
     * has not been explicitly set.
    **/
    public ProxyLog getLog()
    {
	return master;
    }

    /**
     * Returns <code>true</code> if there is a non-<code>null</code>
     * log instance, <code>false</code> otherwise.
    **/
    public boolean hasLog()
    {
	return master != null;
    }

    /**
     * Returns the default parent <code>ClassLoader</code> instance.
    **/
    public ClassLoader getDefaultParentClassLoader()
    {
	return this.parent;
    }

    /**
     * Sets the default parent <code>ClassLoader</code> instance.
    **/
    public synchronized void setDefaultParentClassLoader(ClassLoader parent)
    {
	this.parent = parent;
    }

    /**
     * Returns the default <code>ProxyCompiler</code> instance.
    **/
    public ProxyCompiler getDefaultProxyCompiler()
    {
	return this.compiler;
    }

    /**
     * Sets the default parent <code>ProxyCompiler</code> instance.
    **/
    public synchronized void setDefaultProxyCompiler(ProxyCompiler compiler)
    {
	this.compiler = compiler;
    }

    /**
     * Returns the configuration Properties object.  
    **/
    public Properties getProperties()
    {
	return config;
    }

    /**
     * Returns the <code>Collection</code> of <code>ProxyClass</code>
     * instances in this compiler instance.
     **/
    public synchronized Collection getLoadedClasses()
    {
	return Collections.synchronizedCollection(proxyClasses.values());
    }

    /**
     * JDK1.2 Factory method; Creates a new <code>ProxyClass</code>
     * for the given <code>className</code>.  If a
     * <code>ProxyClass</code> having the given name already exists it
     * will be returned instead.  This is in contrast to the
     * <code>load</code> methods in this class that use dynamic proxy
     * classes.
    **/
    public synchronized ProxyClass loadJDK12(String className)
    {
	ProxyClass pc = getJDK12(className);
	
	if (pc != null) 
	    return pc;

	pc = new JDK12ProxyClass(this, className);
	proxyClasses.put(className, pc);
	return pc;
    }

    /**
     * JDK1.3 Factory method; Creates a new <code>ProxyClass</code>
     * for the given <code>className</code> that will use the
     * <code>JDK1.3</code> dynamic proxy class mechanism.  The
     * ProxyClass will be a factory for <code>Proxy</code> instances
     * having the given interfaces IN ADDITION TO the
     * <code>Proxy</code> interface.  If a <code>ProxyClass</code>
     * having the given name already exists it will be returned
     * instead.
     *
     * <P>If this method is chosen to construct the
     * <code>ProxyClass</code>, new <code>Proxy</code> instances will
     * be assignable to the given interfaces.  Therefore, not only
     * will the return argument of proxyClass.newProxy() be assignable
     * to class <code>Proxy</code>, it will <i>also</i> be assignable
     * to <code>Foo</code>.  Therefore, the following idiom works:
     *
     * <P>
     * <table border=0 cellpadding="0" cellspacing="0" bgcolor="#999999" width="100%">
     * <tr><td>
     * <table cellpadding="0" cellspacing="1"><tr><td><font face="helvetica" color="#eeeeee">
     * Dynamic Proxy Class Programming Idiom
     * </font></td></tr></table>
     * </td></tr>
     * <tr><td>
     * <table border=0 cellpadding="7" cellspacing="1" width="100%"><tr bgcolor="#eeeeff" ><td>
     * <pre>
     * Foo foo = (Foo)loader.load("FooImpl").newInstance();
     * </td></tr></table>
     * </td></tr>
     * </table>
     * <P>
     *
     * This <code>Foo</code> instance will transparently monitor
     * implementation changes and hotswap itself behind the scenes as
     * necessary.
    **/
    public synchronized ProxyClass load(String className, Class[] interfaces)
    {
	ProxyClass pc = getJDK13(className);
	
	if (pc != null) 
	    return pc;
	
	pc = new JDK13ProxyClass(this, className, interfaces);
	proxyClasses.put(className, pc);
	return pc;
    }

    /**
     * JDK1.3 Factory method; Convenience method that uses the
     * published list of interfaces on the Class object.  This is
     * equivalent to <code>load(className,
     * Class.forName(className).getInterfaces())</code>.
    **/
    public synchronized ProxyClass load(String className)
    {
	ProxyClass pc = getJDK13(className);
	
	if (pc != null) 
	    return pc;
	
	pc = new JDK13ProxyClass(this, className);
	proxyClasses.put(className, pc);
	return pc;
    }

    /**
     * Removes the given <code>ProxyClass</code> instance from this
     * cache and returns the cached instance or <code>null</code> if
     * no such <code>ProxyClass</code> exists in this cache.
    **/
    public synchronized ProxyClass remove(ProxyClass cls)
    {
	return (ProxyClass)proxyClasses.remove(cls.getName());
    }
    
    /**
     * Returns the <code>ProxyClass</code> having the given name or
     * <code>null</code> if no such <code>ProxyClass</code> exists; if
     * the given classname is NOT a JDK13 <code>ProxyClass</code>, a
     * <code>ProxyRuntimeException</code> will be thrown.
    **/
    public synchronized ProxyClass getJDK13(String className)
    {
	ProxyClass pc = (ProxyClass)proxyClasses.get(className);
	
	if (pc != null) 
	    if (!pc.isJDK13())
		throw new ProxyRuntimeException
		    ("Attempted fetch of non JDK13ProxyClass "+
		     "through wrong interface.  Use "+
		     "load() instead.");
	    else
		return pc;

	else
	    return null;
    }

    /**
     * Returns the <code>ProxyClass</code> having the given name or
     * <code>null</code> if no such <code>ProxyClass</code> exists; if
     * the given classname is a JDK13 <code>ProxyClass</code>, a
     * <code>ProxyRuntimeException</code> will be thrown.
    **/
    public synchronized ProxyClass getJDK12(String className)
    {
	ProxyClass pc = (ProxyClass)proxyClasses.get(className);
	
	if (pc != null) 
	    if (pc.isJDK13())
		throw new ProxyRuntimeException
		    ("Attempted fetch of JDK13ProxyClass "+
		     "through wrong interface.  Use "+
		     "loadJDK12() instead.");
	    else
		return pc;
	
	else
	    return null;
    }

    // ================================
    // Other Methods 
    // ================================

    PriorityQueue getTimingQueue()
    {
	if (queue == null)
	    queue = new PriorityQueue(this);
	return queue;
    }

    // ================================
    // Debugging Methods 
    // ================================

    protected boolean debug()
    {
	if (lp == 0)
	    return false;

	if (lp < 0) 
	    initChannel();
	
	return (lp & ConsoleLog.PRI_DEBUG) != 0;
    }

    protected void initChannel()
    {
	if (hasLog()) {
	    log = master.newChannel("pcl", this);
	    lp = ConsoleLog.encode(log);
	} else {
	    lp = 0;
	}
    }

    // ================================
    // Instance Fields 
    // ================================

    // A table of ProxyClass objects keyed by className.
    protected Map proxyClasses;

    // The default classloader
    protected ClassLoader parent; 

    // The default compiler
    protected ProxyCompiler compiler; 

    // The default log
    protected ProxyLog master;	

    // The log channel for this class
    protected ProxyLog.Channel log;	

    // The configuration properties
    protected ProxyProperties config;

    // The log priorities
    protected int lp = -1; 

    // The queue for async timing events
    protected PriorityQueue queue;
}
