/**
 * $Id: ProxyCompiler.java,v 1.6 2001/07/27 02:07:00 pcj Exp $
 *
 * Copyright (C) 2001 Paul Cody Johnston - pcj@inxar.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA.
 **/
package org.inxar.hotswap;

import java.io.File;
import java.io.IOException;

import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.LinkedList;
import java.util.Iterator;
import java.util.Properties;
import java.util.StringTokenizer;

import java.util.zip.ZipFile;
import java.util.zip.ZipEntry;
import java.util.jar.JarFile;

/**
 * The <code>ProxyCompiler</code> mediates sourcefile compilation.  
 *
 * <P>
 * <table border=0 cellpadding="0" cellspacing="0" bgcolor="#999999" width="100%">
 * <tr><td>
 * <table cellpadding="0" cellspacing="1"><tr><td><font face="helvetica" color="#eeeeee">
 * ProxyCompiler Usage
 * </font></td></tr></table>
 * </td></tr>
 * <tr><td>
 * <table border=0 cellpadding="7" cellspacing="1" width="100%"><tr bgcolor="#eeeeff" ><td>
 * <pre>
 * ProxyCompiler compiler = new JavacCompiler();
 * ProxyCompiler compiler = new SystemCompiler("jikes");
 *
 * compiler.setDestinationpath("path/to/location/where/classfiles/are/written/");
 * compiler.setSourcepath("path/to/location/where/sourcefiles/are/searched/");
 * compiler.getClasspath().add("path/to/any/other/necessary/libraries/or/jarfiles");
 *
 * proxyClass.setCompiler(compiler);
 * proxyClassLoader.setDefaultCompiler(compiler);
 * </td></tr></table>
 * </td></tr>
 * </table>
 * <P>
 *
 * @since 0.8.2
**/
public abstract class ProxyCompiler implements PriorityQueue.Listener
{
    private static final boolean LOG = true;

    /**
     * <code>int getStatus</code> return code signifying that a
     * the Class object in memory is up to date.
     **/
    public static final int STATUS_CURRENT = 0;

    /**
     * <code>int getStatus</code> return code signifying that the
     * object in memory is out of date with the classfile, it needs to
     * be reloaded from the classfile.
     **/
    public static final int STATUS_LOAD = 1;

    /**
     * <code>int getStatus</code> return code signifying the classfile
     * is out of date and needs to be recompiled.  Also returned when
     * a sourcefile exists but no classfile has yet been compiled.
     **/
    public static final int STATUS_COMPILE = 2;

    /**
     * <code>int getStatus</code> return code signifying that no
     * classfile or sourcefile could be found EVEN though the object
     * was previously loaded into memory (suggesting that both the
     * classfile and sourcefile were subsequently deleted).
     **/
    public static final int STATUS_WARN = -1;

    /**
     * <code>int getStatus</code> return code signifying that the
     * class needs to be loaded, but no classfile or sourcefile could
     * be found.
     **/
    public static final int STATUS_ERROR = -2;

    /**
     * <code>int getStatus</code> return code signifying that a the
     * status could not be determined.
     **/
    public static final int STATUS_UNKNOWN = -3;

    /**
     * <code>int compile</code> return code signifying that a
     * compilation was attempted and completed successfully.
     **/
    public static final int COMPILE_SUCCESS = 0;

    /**
     * <code>int compile</code> return code signifying that a
     * compilation was attempted and NOT completed successfully.
     **/
    public static final int COMPILE_FAILURE = 1;

    /**
     * <code>int compile</code> return code signifying the outcome of
     * a compilation could not be determined.
     **/
    public static final int COMPILE_UNKNOWN = -1;

    /**
     * Cached instance of
     * <code>System.getProperty("path.separator")</code> (unix = ':',
     * dos = ';').
     **/
    static String PATHSEP = System.getProperty("path.separator");

    // ================================
    // Constructors 
    // ================================

    protected ProxyCompiler()
    {
	init();
    }

    protected void init()
    {
	this.map = new HashMap();
    }

    // ================================
    // TEHC Methods
    // ================================

    // ================================
    // Abstract Methods 
    // ================================

    /**
     * Compiles the class associated with the given
     * <code>ProxyClass</code>.
     **/
    public abstract ProxyCompileEvent compile(ProxyClass cls);

    // ================================
    // PriorityQueue.Listener Methods 
    // ================================

    /**
     * Asynchronous callback method that is invoked by the priority
     * queue thread.
    **/
    public void queue_notify(Object callback)
    {
	wasTripped = true;
    }

    // ================================
    // Public Methods 
    // ================================

    /**
     * Sets the parent <code>ProxyClassLoader</code>.  This must be
     * set directly after instantiation of the
     * <code>ProxyCompiler</code>.
    **/
    public void setProxyClassLoader(ProxyClassLoader loader)
    {
	this.loader = loader;
    }

    /**
     * Initializes the <code>ProxyCompiler</code> instance from the
     * given <code>Properties</code>.  If initialization was
     * unsuccessful, return <code>false</code>.
    **/
    public boolean initialize(Properties p)
    {
	Object o;
	String s;
 
	o = p.get("compiler.sourcepath");
	if (o instanceof String)
	    setSourcepath(o);

	o = p.get("compiler.destinationpath");
	if (o instanceof String) 
	    setDestinationpath((String)o);

	o = p.get("compiler.classpath");
	if (o instanceof String) 
	    getClasspath().add(o);
	else if (o instanceof List) 
	    getClasspath().addAll((List)o);

	o = p.get("compiler.option");
	if (o instanceof String) 
	    getOptions().add(o);
	else if (o instanceof List) 
	    getOptions().addAll((List)o);

	s = ProxyProperties.getFirstProperty("compiler.usejavaclasspath", p);
	if (s != null && new Boolean(s).booleanValue()) {
	    s = System.getProperty("java.class.path");
	    if (s != null) {
		StringTokenizer t = new StringTokenizer(s, File.pathSeparator);
		while (t.hasMoreTokens())
		    getClasspath().add(t.nextToken());
	    }
	}

	s = ProxyProperties.getFirstProperty("compiler.asynchronous", p);
	if (s != null) {
	    if (new Boolean(s).booleanValue()) {
		s = ProxyProperties.getFirstProperty("compiler.asyncinterval", p);
		if (s == null)
		    throw new ProxyRuntimeException
			("compiler.asyncinterval is required in asynchronous mode");
		try {
		    asyncInterval = Long.parseLong(s);
		} catch (NumberFormatException nfex) {
		    throw new ProxyRuntimeException(nfex);
		} 
		isAsync = true;
		wasTripped = false;
		PriorityQueue queue = loader.getTimingQueue();
		queue.schedule(this, null, asyncInterval);
	    }
	}

	if (debug())
	    initDebug();

	return true;
    }

    protected void initDebug()
    {
	log.debug("<init> asynchronous mode? "+isAsync);
	if (isAsync)
	    log.debug("<init> asynchronous interval: "+asyncInterval);
	log.debug("<init> sourcepath: "+getSourcepath());
	log.debug("<init> destinationpath: "+getDestinationpath());
	log.debug("<init> options: "+getOptions());
	log.debug("<init> classpath: "+getClasspath());
    }

    /**
     * Sets the <code>Object</code> where the source files packages
     * are located.  The run-time type of this <code>Object</code> may
     * be a <code>String</code> or <code>File</code> (may add
     * <code>URL</code> later).
     **/
    public synchronized void setSourcepath(Object sourcepath)
    {
	if (sourcepath == null || 
	    sourcepath instanceof String || 
	    sourcepath instanceof File)
	    this.sourcepath = sourcepath;

	else
	    throw new IllegalArgumentException
		("Sourcepath must be a String, File, or null.");
    }

    /**
     * Returns the location where the source files packages are
     * located.
     **/
    public Object getSourcepath()
    {
	return this.sourcepath;
    }

    /**
     * Returns the pathname where the compiled classes will be
     * written.
     **/
    public synchronized void setDestinationpath(String dst)
    {
	this.dst = dst;
    }

    /**
     * Sets the pathname where the class files should be written (and
     * read from).
     **/
    public String getDestinationpath()
    {
	return this.dst;
    }

    /**
     * Returns the <code>List</code> of the option entries.  Users
     * should interact with the <code>List</code> directly to add
     * <code>Object</code> instances whose <code>toString()</code>
     * method will render appropriately.  Each entry is printed as-is,
     * no dashes are added by the compiler.
     **/
    public synchronized List getOptions()
    {
	if (opts == null) 
	    opts = new LinkedList();
	return opts;
    }

    /**
     * Returns the <code>List</code> of the classpath entries.  Users
     * should interact with the <code>List</code> directly to add
     * <code>Object</code> instances whose <code>toString()</code>
     * method will render appropriately.
     **/
    public synchronized List getClasspath()
    {
	if (cps == null) 
	    cps = new LinkedList();
	return cps;
    }

    /**
     * Adds the given listener of <code>ProxyCompileEvent</code>
     * notifications.
     **/
    public synchronized void addCompileListener(ProxyEventListener l)
    {
	if (listeners == null)
	    listeners = new LinkedList();
	listeners.add(l);
    }

    /**
     * Removes the given listener of <code>ProxyCompileEvent</code>
     * notifications.
     **/
    public synchronized void removeCompileListener(ProxyEventListener l)
    {
	if (listeners != null) 
	    listeners.remove(l);
    }

    // ================================
    // Package Methods 
    // ================================

    /**
     * Returns the sourcefile <code>ProxyResource</code> for this
     * <code>ProxyClass</code>.
    **/
    public synchronized ProxyResource getSourcefile(ProxyClass cls)
    {
	return lookup(cls).getSourcefile();
    }

    /**
     * Returns the classfile <code>ProxyResource</code> for this
     * <code>ProxyClass</code>.
    **/
    public synchronized ProxyResource getClassfile(ProxyClass cls)
    {
	return lookup(cls).getClassfile();
    }

    /**
     * Returns the sourcefile <code>Resource</code> for the given
     * classname.  The location of the sourcecode is dependent upon
     * the run-time type of the sourcepath <code>Object</code> (see
     * <code>setSourcepath(Object)</code>).
     **/
    synchronized ProxyResource getSourcefile(String className)
    {
	StringBuffer b = new StringBuffer();
	
	if (sourcepath != null) {
	    if (sourcepath instanceof String) {
		String s = (String)sourcepath;
		if (s.startsWith("file:"))
		    b.append(s.substring(5));
		else 
		    b.append(s);
	    } else if (sourcepath instanceof File) {
		File f = (File)sourcepath;
		b.append(f.getAbsolutePath());
	    }		
	}
	
	if (b.length() > 0 && b.charAt(b.length() - 1) != File.separatorChar)
	    b.append(File.separatorChar);
	
	b.append(className.replace('.' , File.separatorChar))
	    .append(".java");
	
	if (debug())
	    log.debug("<getSourcefile> Sourcefile for "+className+
		      " believed to be file:"+b.toString());

	return new FileResource(new File(b.toString()));
    }

    /**
     * Returns a File resource for the given classname.  It is always
     * assumed that the classfile exists on the filesystem.
     **/
    synchronized ProxyResource getClassfile(String className)
    {
	    StringBuffer b = new StringBuffer();

	    if (dst != null) {
		b.append(dst);
		if (dst.charAt(dst.length() - 1) != File.separatorChar)
		    b.append(File.separatorChar);
	    }

	    b.append(className.replace('.' , File.separatorChar))
		.append(".class");

	    ProxyResource rsc = new FileResource(new File(b.toString()));

	    // Now we'll check to see if this resource exists.  If it does
	    // NOT exist, we will look through the classpath and see if
	    // any jar or zip files are there, and systematically try to
	    // find a corresponding entry in the archive.
	    try {
	    
		if (!rsc.exists() && cps != null && cps.size() > 0) {

		    // Try to find it in a zip or jar file in the classpath.
		    // First thing is to reformat the filename NOT relative to
		    // the dst path.
		    String classFileName = 
			className.replace('.' , File.separatorChar) + ".class";
		
		    Iterator i = cps.iterator();
		
		    Object o;
		    ZipFile arch;
		
		    while (i.hasNext()) {
			o = i.next();
		    
			// Assign the archive variable.
			if (o instanceof String) {
			    String cp = (String)o;
			    if (cp.endsWith(".zip"))
				if (!new File(cp).exists()) {
				    if (warn())
					log.warn("Cannot find file:"+cp);
				    continue;
				} else {
				    arch = new ZipFile(cp);
				}
			    else if (cp.endsWith(".jar")) {
				// Ignore two well known jars
				if (cp.endsWith("rt.jar") || 
				    cp.endsWith("tools.jar"))
				    continue;
				// Okay, we think this is a 3rd party
				// jar, make sure the jar exists.
				if (!new File(cp).exists()) {
				    if (warn())
					log.warn("Cannot find file:"+cp);
				    continue;
				} else {
				    arch = new JarFile(cp);
				}
			    } else {
				// skip this classpath entry, it does not
				// look like a zip or jar file.
				continue;
			    }
			} else if (o instanceof ZipFile) {
			    arch = (ZipFile)o;
			} else {
			    // unknown classpath entry type.  Skip it.
			    continue;
			}
		    
			// OK, Query the archive to see if the
			// file is there.
			ZipEntry e = arch.getEntry(classFileName);
		    
			if (e != null) {
			    // Success!  Looks like we found it.  Replace
			    // the string classpath entry with the zipfile
			    // one for future searches and less memory
			    // consumption (or more?)  TESTME.
			    if ( ! (o instanceof ZipFile) ) {
				i.remove();
				cps.add(arch);
				/* Won't this throw a concurrent
				   modification exception? (Not unless
				   you actually *use* the iterator
				   again, which you do not) */
			    }
			
			    rsc =  new ZipResource(arch, e);

			    if (debug())
				log.debug("<getClassfile> Classfile for "+className+
					  " believed to be "+rsc.toString());

			    // Return to caller.
			    return rsc;
			}
		    }
		}

	    } catch (IOException ioex) {
		ioex.printStackTrace();
		/* Should not happen */
	    } 

	    /* NOTE: One case where this fails is that there are more than
	       one classfile entries in the classpath for this classname,
	       one in a zip or jar file, and one that DOES NOT YET exist
	       but presumably WOULD if we compiled it from source.  In
	       this case, we are setting up the classfile resource from
	       the jar file, and hence will NEVER compile it.  The
	       expectation here is that the application is dynamically
	       updating jar files (like jboss, for example), but not
	       source code.  Thus, we want to reload the new classes from
	       the jar, knowing that we'll never be recompiling. */

	    if (debug())
		log.debug("<getClassfile> Classfile for "+className+
			  " believed to be file:"+rsc.toString());

	    return rsc;
    }

    /**
     * Returns the status of <code>ProxyClass</code> as one of the
     * <code>STATUS_XXX</code> constants in this class.
     **/
    public synchronized int getStatus(ProxyClass proxyClass)
    {
	if (isAsync) {
	    if (wasTripped) {
		if (debug()) 
		    log.debug("<status> asynchronous trigger pulled"+
			      ", proceeding with non-naiive status check.");
		wasTripped = false;
	    } else {
		if (debug()) 
		    log.debug("<status> "+proxyClass.getName()+
			      " is in asynchronous mode; naiively reporting status current.");
		return STATUS_CURRENT;
	    }
	}

	try {

	    Entry e = lookup(proxyClass);
	    ProxyResource src = e.getSourcefile();
	    ProxyResource cls = e.getClassfile();
	    long obj = proxyClass.getLastClassload();

	    if (src.exists()) {
		if (cls.exists()) {
		    if (src.lastModified() > cls.lastModified()) {
			if (debug()) 
			    log.debug("<status> "+proxyClass.getName()+" needs compile (s & c & s>c)");
			return STATUS_COMPILE;
		    } else if (cls.lastModified() > obj) {
			if (debug()) 
			    log.debug("<status> "+proxyClass.getName()+" needs classload (s & c & c>o, c>s)");
			return STATUS_LOAD;
		    } else {
			if (debug()) 
			    log.debug("<status> "+proxyClass.getName()+" is current (s & c & o>c)");
			return STATUS_CURRENT;
		    }
		} else {
		    if (debug()) 
			log.debug("<status> "+proxyClass.getName()+" needs compile (s & !c)");
		    return STATUS_COMPILE;
		}

	    } else if (cls.exists()) {
		if (cls.lastModified() > obj) {
		    if (warn()) 
			log.warn("<status> "+proxyClass.getName()+" sourcefile is absent (!s & c & c>o)");
		    if (debug()) 
			log.debug("<status> "+proxyClass.getName()+" needs classload (!s & c & c>o)");
		    return STATUS_LOAD;
		} else {
		    if (warn()) 
			log.warn("<status> "+proxyClass.getName()+" sourcefile is absent (!s & o>c)");
		    if (debug()) 
			log.debug("<status> "+proxyClass.getName()+" is current (!s & o>c)");
		    return STATUS_CURRENT; 
		}

	    } else if (obj > 0) {
		if (warn()) 
		    log.warn("<status> "+proxyClass.getName()+" is stranded! (!s & !c)");
		return STATUS_WARN;
	    } else {
		if (fatal()) 
		    log.fatal("<status> "+proxyClass.getName()+" is nonexistent! (!s & !c & !o)");
		return STATUS_ERROR;
	    }

	} catch (IOException ioex) {

	    ioex.printStackTrace();
	    if (debug()) 
		log.debug("<status> "+proxyClass.getName()+" unknown ("+ioex.getMessage()+")");
	    return STATUS_UNKNOWN;

	} 
    }

    // ================================
    // Protected Methods 
    // ================================

    protected Entry lookup(ProxyClass cls)
    {
	Entry e = (Entry)map.get(cls);
	if (e == null) {
	    e = this.new Entry(cls);
	    map.put(cls, e);
	}
	return e;
    }

    /**
     * Sends <code>ProxyRecompileEvent</code> notification to all
     * listeners.
     **/
    protected synchronized void fire(ProxyCompileEvent evt)
    {
	if (hasListeners()) {
	    Iterator i = listeners.iterator();
	    while (i.hasNext()) 
		((ProxyEventListener)i.next()).notify(evt);
	}
    }

    /**
     * Returns <code>true</code> if there is at least one listener.
     **/
    protected synchronized boolean hasListeners()
    {
	return listeners != null && listeners.size() > 0;
    }

    // ================================
    // Debugging Methods 
    // ================================

    protected boolean fatal()
    {
	if (lp == 0)
	    return false;

	if (lp < 0) 
	    initChannel();
	
	return (lp & ConsoleLog.PRI_FATAL) != 0;
    }

    protected boolean warn()
    {
	if (lp == 0)
	    return false;

	if (lp < 0) 
	    initChannel();
	
	return (lp & ConsoleLog.PRI_WARN) != 0;
    }

    protected boolean debug()
    {
	if (lp == 0)
	    return false;

	if (lp < 0) 
	    initChannel();
	
	return (lp & ConsoleLog.PRI_DEBUG) != 0;
    }

    protected void initChannel()
    {
	if (loader.hasLog()) {
	    log = loader.getLog().newChannel("pcc", this);
	    lp = ConsoleLog.encode(log);
	} else {
	    lp = 0;
	}
    }

    // ================================
    // Instance Fields 
    // ================================

    protected Map map;		// map of Entry objects
    protected List cps;		// classpaths
    protected List opts;	// options
    protected String dst;	// code destination
    protected Object sourcepath; // the sourcepath
    protected List listeners;	// event listeners
    protected ProxyClassLoader loader;
    protected boolean isAsync;
    protected boolean wasTripped;
    protected long asyncInterval;
    protected ProxyLog.Channel log;	
    protected int lp = -1; 

    protected class Entry
    {
	Entry(ProxyClass cls)
	{
	    init(cls);
	}

	void init(ProxyClass cls)
	{
	    String className = cls.getName();
	    this.src = ProxyCompiler.this.getSourcefile(className);
	    this.cls = ProxyCompiler.this.getClassfile(className);
	}

	ProxyResource getSourcefile()
	{
	    return src;
	}

	ProxyResource getClassfile()
	{
	    return cls;
	}	

	ProxyResource src;
	ProxyResource cls;
    }
}

