/**
 * $Id: SystemCompiler.java,v 1.1 2001/07/27 02:07:00 pcj Exp $
 *
 * Copyright (C) 2001 Paul Cody Johnston - pcj@inxar.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA.
 **/
package org.inxar.hotswap;

import java.io.File;
import java.io.StringWriter;
import java.io.PrintWriter;
import java.io.BufferedInputStream;
import java.io.IOException;

import java.util.List;
import java.util.Iterator;
import java.util.Properties;

/**
 * Concrete <code>ProxyCompiler</code> implementation that executes a
 * compilation command through <code>Runtime.exec</code>.
 *
 * <P>
 * <table border=0 cellpadding="0" cellspacing="0" bgcolor="#999999" width="100%">
 * <tr><td>
 * <table cellpadding="0" cellspacing="1"><tr><td><font face="helvetica" color="#eeeeee">
 * Constructor usage of SystemCompiler
 * </font></td></tr></table>
 * </td></tr>
 * <tr><td>
 * <table border=0 cellpadding="7" cellspacing="1" width="100%"><tr bgcolor="#eeeeff" ><td>
 * <pre>
 * ProxyCompiler compiler = new SystemCompiler();
 * compiler.setProxyClassLoader(loader);
 *
 * compiler.setCommandName("javac");
 * compiler.setCommandName("jikes");
 * compiler.setCommandName("/usr/local/jdk1.2.2/javac");
 * compiler.setCommandName("/usr/bin/jikes");
 * </td></tr></table>
 * </td></tr>
 * </table>
 * <P>
 *
 * @since 0.8.2, renamed in 0.8.6
**/
public class SystemCompiler extends ProxyCompiler
{
    // ================================
    // Constructors 
    // ================================

    /**
     * Constructs a new <code>ProxyCompiler</code> having a default
     * <code>ClassLoader</code>.  
     **/
    public SystemCompiler()
    {
	super();
    }

    public boolean initialize(Properties p)
    {
	// Make sure superclass initializes okay.
	if (!super.initialize(p))
	    return false;

	// Fetch the command object.  It may be a string or a list
	// depending on how many were provided.
	Object o = p.get("compiler.command");
	if (o instanceof String) {
	    commandName = (String)o;
	    if (!canExecute())
		return false;
	} else if (o instanceof List) {
	    Iterator i = ((List)o).iterator();
	    while (i.hasNext()) {
		commandName = (String)i.next();
		if (canExecute())
		    break;
	    }
	} else {
	    // The command name must have been null which is
	    // unacceptable.
	    return false;
	}

	if (commandName.endsWith("jikes")) {
	    String home = System.getProperty("java.home");
	    if (home != null) {
		if (home.endsWith("/jre")) {
		    int len = home.length() - 4;
		    home = home.substring(0, len);
		}			
		
		getClasspath().add(home + "/jre/lib/rt.jar");
		getClasspath().add(home + "/lib/tools.jar");
	    }
	}

	if (debug())
	    myInitDebug();

	return true;
    }

    protected void initDebug()
    {
	/* override to say nothing, we'll use te myInitDebug method
           instead. */
    }

    protected void myInitDebug()
    {
	log.debug("<init> command: "+commandName);
	log.debug("<init> sourcepath: "+getSourcepath());
	log.debug("<init> destinationpath: "+getDestinationpath());
	log.debug("<init> options: "+getOptions());
	log.debug("<init> classpath: "+getClasspath());
    }

    /**
     * Used to check that the command name is valid and that it can
     * indeed be run on the system.
    **/
    protected boolean canExecute()
    {
	if (commandName == null)
	    return false;

	// The way we will determine if a command is executable is by
	// just running the named command in naked form, without
	// arguments.  We assume that the output will have at least a
	// minimum amount of output, say for example 50 characters
	// plus the size of the command itself.  This output we assume
	// to be the "usage" content.
	try {

	    SystemCommand cmd = new SystemCommand(commandName);
	    cmd.reportIOExceptions(false);
	    cmd.run();
	    
	    if (cmd.err != null && cmd.err.length() > commandName.length() + 50)
		return true;
	    
	    if (cmd.out != null && cmd.out.length() > commandName.length() + 50)
		return true;
	    
	} catch (Exception ex) {
	    if (ex instanceof IOException)
		if (debug())
		    log.debug("<init> Looks like a bad command name: "+commandName+
			      " ("+ex.getMessage()+")");

	} 

	return false;
    }

    // ================================
    // TEHC Methods
    // ================================

    // ================================
    // Public Methods 
    // ================================

    /**
     * Sets the command name of the system compiler, for example
     * "<code>/usr/bin/jikes</code>" or "<code>javac</code>".
    **/
    public void setCommandName(String commandName)
    {
	this.commandName = commandName;
    }

    // ================================
    // ProxyCompiler Methods 
    // ================================

    public synchronized ProxyCompileEvent compile(ProxyClass cls) 
    {
	Entry e = lookup(cls);
	String className = cls.getName();
	ProxyResource sourceFile = e.getSourcefile();
	ProxyResource classFile = e.getClassfile();

	boolean wasSuccessful = false;
	String out = "";
	String err = "";

	StringBuffer cmd = getCommand()
	    .append(' ')
	    .append(sourceFile.toString());

	SystemCommand c = null;

	try {

	    if (classFile.exists()) {

		long lastMod = classFile.lastModified();
		
		c = new SystemCommand(cmd.toString());
		c.run();
		
		wasSuccessful = lastMod < classFile.lastModified();
		
	    } else {
		
		c = new SystemCommand(cmd.toString());
		c.run();
		
		wasSuccessful = classFile.exists();
		
	    }

	} catch (IOException ioex) {

	    StringWriter sw = new StringWriter();
	    PrintWriter pw = new PrintWriter(sw);
	    ioex.printStackTrace(pw);
	    err = sw.toString();
	    wasSuccessful = false;

	}

	if (c != null) {
	    out += c.out;
	    err += c.err;
	}

	int rc = wasSuccessful ? COMPILE_SUCCESS : COMPILE_FAILURE;

	ProxyCompileEvent evt = new ProxyCompileEvent
	    (this, className, cmd.toString(), out, err, rc);

	if (debug())
	    log.debug(evt.toString());

	if (hasListeners())
	    fire(evt);

	return evt;
    }

    // ================================
    // Other Methods 
    // ================================

    protected StringBuffer getCommand() 
    {
	StringBuffer cmd = new StringBuffer(commandName);
	
	if (opts != null) {
	    int count = 0;
	    Iterator i = opts.iterator();
	    while (i.hasNext()) {
		cmd.append(' ').append(i.next());
	    }
	}

	if (dst != null)
	    cmd.append(" -d ").append(dst);
	
//  	if (sourcepath != null)
//  	    cmd.append(" -sourcepath ").append(sourcepath);
	
	if (cps != null) {
	    cmd.append(" -classpath ");
	    int count = 0;
	    Iterator i = cps.iterator();
	    while (i.hasNext()) {
		if (count++ > 0)
		    cmd.append(PATHSEP);
		cmd.append(i.next());
	    }
	}

	return cmd;
    }

    // ================================
    // Fields 
    // ================================

    // compiler command name (jikes, javac)
    protected String commandName;	
}
